package com.ibm.tools.attach.javaSE;

/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 2009, 2011  All Rights Reserved
 */

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.util.List;
import java.util.Properties;

import com.ibm.tools.attach.AgentInitializationException;
import com.ibm.tools.attach.AgentLoadException;
import com.ibm.tools.attach.AttachNotSupportedException;
import com.ibm.tools.attach.VirtualMachineDescriptor;
import com.ibm.tools.attach.spi.AttachProvider;

/**
 * Handles the initiator end of an attachment to a target VM
 *
 */
class VirtualMachineImpl extends com.ibm.tools.attach.VirtualMachine {

	static final String COM_IBM_TOOLS_ATTACH_TIMEOUT = "com.ibm.tools.attach.timeout"; //$NON-NLS-1$
	static final String COM_IBM_TOOLS_COMMAND_TIMEOUT = "com.ibm.tools.attach.command_timeout"; //$NON-NLS-1$
	/* The units for timeouts are milliseconds, Set to 0 for no timeout. */
	private static final int DEFAULT_ATTACH_TIMEOUT = 120000;	/* should be ~2* the TCP timeout, i.e. /proc/sys/net/ipv4/tcp_fin_timeout on Linux */
	private static final int DEFAULT_COMMAND_TIMEOUT = 0;

	private static int ATTACH_TIMEOUT = Integer.getInteger(COM_IBM_TOOLS_ATTACH_TIMEOUT, DEFAULT_ATTACH_TIMEOUT).intValue();
	private static int COMMAND_TIMEOUT = Integer.getInteger(COM_IBM_TOOLS_COMMAND_TIMEOUT, DEFAULT_COMMAND_TIMEOUT).intValue();

	private static final String INSTRUMENT_LIBRARY = "instrument"; //$NON-NLS-1$
	private static final int ATTACH_RETRY_LIMIT = 4;

	private OutputStream commandStream;
	private final VirtualMachineDescriptorImpl descriptor;
	private final AttachProviderImpl myProvider;
	private Integer portNumber;
	private InputStream responseStream;
	private boolean targetAttached;
	private String targetId;
	private FileLock[] targetLocks;
	private ServerSocket targetServer;
	private Socket targetSocket;

	/**
	 * @param provider
	 *            AttachProvider which created this VirtualMachine
	 * @param id
	 *            identifier for the VM
	 */
	VirtualMachineImpl(AttachProvider provider, String id)
			throws NullPointerException {
		super(provider, id);
		if ((null == id) || (null == provider)) {
			// K0554 = Virtual machine ID or display name is null
			throw new NullPointerException(com.ibm.oti.util.Msg.getString("K0554")); //$NON-NLS-1$
		}
		new IPC();
		this.targetId = id;
		this.myProvider = (AttachProviderImpl) provider;
		this.descriptor = (VirtualMachineDescriptorImpl) myProvider.getDescriptor(id);
		if (null != descriptor) {
			CommonDirectory.setSemaphoreName(descriptor.getNotifierId());
		}
	}

	/**
	 * @throws IOException
	 *             if cannot open the communication files
	 * @throws AttachNotSupportedException
	 *             if the descriptor is null the target does not respond.
	 */
	void attachTarget() throws IOException, AttachNotSupportedException {
		AttachNotSupportedException lastException = null;
		for (int i = 0; i < ATTACH_RETRY_LIMIT; ++i) {
			lastException = null;
			try {
				tryAttachTarget();
			} catch (AttachNotSupportedException e) {
				IPC.logMessage("attachTarget " + targetId + " timeout #" + i); //$NON-NLS-1$ //$NON-NLS-2$
				lastException = e;
			}
			if (null == lastException) {
				break;
			}
		}
		if (null != lastException) {
			throw lastException;
		}
	}

	private String createLoadAgent(String agentName, String options) {
		String optString = (null == options) ? "" : //$NON-NLS-1$
				options;
		String cmdFrame = Command.LOADAGENT + '(' + INSTRUMENT_LIBRARY + ','
				+ agentName + '=' + optString + ')';
		return cmdFrame;
	}

	private String createLoadAgentLibrary(String agentName, String options,
			boolean agentPath) {
		String cmd = (agentPath ? Command.LOADAGENTPATH
				: Command.LOADAGENTLIBRARY);
		String optString = ""; //$NON-NLS-1$
		if ((null != options) && !(options.equals(""))) { //$NON-NLS-1$
			optString = ',' + options;
		}
		String cmdFrame = cmd + '(' + agentName + optString + ')';
		return cmdFrame;
	}

	/**
	 * Detach a virtual machine by sending a detach command to the connected VM
	 * via the IPC channel.
	 */
	@Override
	public synchronized void detach() throws IOException {
		AttachmentConnection.streamSend(commandStream, Command.DETACH);
		try {
			AttachmentConnection.streamReceiveString(responseStream);
		} finally {
			IPC.logMessage("VirtualMachine.detach"); //$NON-NLS-1$
			if (null != commandStream) {
				commandStream.close();
				commandStream = null;
			}
			if (null != targetSocket) {
				targetSocket.close();
				targetSocket = null;
			}
			if (null != targetServer) {
				targetServer.close();
				targetServer = null;
			}
		}
		targetAttached = false;
	}

	@Override
	public Properties getAgentProperties() throws IOException {
		if (!targetAttached) {
			// K0544 = Target not attached
			throw new IOException(com.ibm.oti.util.Msg.getString("K0544")); //$NON-NLS-1$
		}
		Properties props = getTargetProperties(false);

		return props;
	}

	@Override
	public Properties getSystemProperties() throws IOException {
		if (!targetAttached) {
			// K0544 = Target not attached
			throw new IOException(com.ibm.oti.util.Msg.getString("K0544")); //$NON-NLS-1$
		}
		Properties props = getTargetProperties(true);
		return props;
	}

	/**
	 * Synchronized because multiple threads may use this object
	 * @param systemProperties flag to indicate if we want system properties (true) or agent properties (false)
	 * @return Properties object
	 * @throws IOException
	 */
	private synchronized Properties getTargetProperties(boolean systemProperties)
			throws IOException {
		AttachmentConnection.streamSend(commandStream,
				systemProperties ? Command.GET_SYSTEM_PROPERTIES
						: Command.GET_AGENT_PROPERTIES);
		Properties props = new Properties();
		byte msgBuff[] = AttachmentConnection.streamReceiveBytes(responseStream);

		props.load(new ByteArrayInputStream(msgBuff));
		return props;
	}

	@Override
	public synchronized void loadAgent(String agent, String options)
			throws AgentLoadException, AgentInitializationException,
			IOException {

		if (!targetAttached) {
			// K0544 = Target not attached
			throw new IOException(com.ibm.oti.util.Msg.getString("K0544")); //$NON-NLS-1$
		}
		AttachmentConnection.streamSend(commandStream, (createLoadAgent(agent, options)));
		String response = AttachmentConnection.streamReceiveString(responseStream);
		parseResponse(response);
	}

	@Override
	public synchronized void loadAgentLibrary(String agentLibrary,
			String options) throws AgentLoadException,
			AgentInitializationException, IOException {

		if (!targetAttached) {
			// K0544 = Target not attached
			throw new IOException(com.ibm.oti.util.Msg.getString("K0544")); //$NON-NLS-1$
		}
		AttachmentConnection.streamSend(commandStream, createLoadAgentLibrary(
				agentLibrary, options, false));
		String response = AttachmentConnection.streamReceiveString(responseStream);
		parseResponse(response);
	}

	@Override
	public synchronized void loadAgentPath(String agentPath, String options)
			throws AgentLoadException, AgentInitializationException,
			IOException {
		if (!targetAttached) {
			// K0544 = Target not attached
			throw new IOException(com.ibm.oti.util.Msg.getString("K0544")); //$NON-NLS-1$
		}
		AttachmentConnection.streamSend(commandStream, createLoadAgentLibrary(agentPath,
				options, true));
		String response = AttachmentConnection.streamReceiveString(responseStream);
		parseResponse(response);
	}

	private void lockAllAttachNotificationSyncFiles(
			List<VirtualMachineDescriptor> vmds) {

		int vmdIndex = 0;
		targetLocks = new FileLock[vmds.size()];
		for  (VirtualMachineDescriptor i : vmds) {
			VirtualMachineDescriptorImpl vmd;
			try {
				vmd = (VirtualMachineDescriptorImpl) i;
			} catch (ClassCastException e) {
				continue;
			}
			if (!vmd.id().equalsIgnoreCase(AttachHandler.getVmId())) { /*
																		 * avoid
																		 * overlapping
																		 * locks
																		 */
				String attachSyncFile = vmd.getAttachSyncFileValue();
				if (null != attachSyncFile) { /*
				 * in case of a malformed advert
				 * file
				 */
					IPC.logMessage("lockAllAttachNotificationSyncFiles locking targetLocks[", vmdIndex, "] ", attachSyncFile); //$NON-NLS-1$ //$NON-NLS-2$
					targetLocks[vmdIndex] = new FileLock(attachSyncFile, TargetDirectory.SYNC_FILE_PERMISSIONS);
					try {
						targetLocks[vmdIndex].lockFile(true);
					} catch (IOException e) {
						targetLocks[vmdIndex] = null;
						IPC.logMessage("lockAllAttachNotificationSyncFiles locking targetLocks[", vmdIndex, "] ", "already locked"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
					}
				}
			} else {
				targetLocks[vmdIndex] = null;
			}
			++vmdIndex;
		}
	}

	private void parseResponse(String response) throws IOException,
			AgentInitializationException, AgentLoadException {
		if (response.startsWith(Response.ERROR.toString())) {
			int responseLength = response.indexOf('\0');
			String trimmedResponse;
			if (-1 == responseLength) {
				trimmedResponse = response;
			} else {
				trimmedResponse = response.substring(0, responseLength);
			}
			if (response.contains(Response.EXCEPTION_IOEXCEPTION)) {
				throw new IOException();
			} else if (response
					.contains(Response.EXCEPTION_AGENT_INITIALIZATION_EXCEPTION)) {
				throw new AgentInitializationException(trimmedResponse);
			} else if (response
					.contains(Response.EXCEPTION_AGENT_LOAD_EXCEPTION)) {
				throw new AgentLoadException(trimmedResponse);
			}
		}

	}

	private void tryAttachTarget() throws IOException,
			AttachNotSupportedException {
		Reply replyFile = null;
		AttachHandler.waitForAttachApiInitialization(); /* ignore result: we can still attach to another target if API is disabled */
		IPC.logMessage("VirtualMachineImpl.tryAttachtarget"); //$NON-NLS-1$
		Object myIn = AttachHandler.getMainHandler().getIgnoreNotification();

		synchronized (myIn) {
			int numberOfTargets = 0;
			try {
				CommonDirectory.obtainAttachLock();
				if (null == descriptor) {
					// K0531 = target not found
					throw new AttachNotSupportedException(com.ibm.oti.util.Msg.getString("K0531")); //$NON-NLS-1$
				}
				List<VirtualMachineDescriptor> vmds = myProvider
						.listVirtualMachines();
				if (null == vmds) {
					return;
				}

				targetServer = new ServerSocket(0); /* select a free port */
				portNumber = Integer.valueOf(targetServer.getLocalPort());
				String key = Integer.toHexString((IPC.getRandomNumber()));
				replyFile = new Reply(portNumber, key, TargetDirectory.getTargetDirectoryPath(descriptor.id()), descriptor.getUid());
				try {
					replyFile.writeReply();
				} catch (IOException e) { /*
										 * target shut down while we were trying
										 * to attach
										 */
					// K0457 = Target no longer available
					throw new AttachNotSupportedException(com.ibm.oti.util.Msg.getString("K0457")); //$NON-NLS-1$
				}

				if (descriptor.id().equals(AttachHandler.getVmId())) {
					/* I am connecting to myself: bypass the notification and launch the attachment thread directly */
					if (AttachHandler.waitForAttachApiInitialization()) {
						AttachHandler.getMainHandler().connectToAttacher();
					} else {
						// K0558 = Attach API initialization failed
						throw new AttachNotSupportedException(com.ibm.oti.util.Msg.getString("K0558")); //$NON-NLS-1$
					}
				} else {
					lockAllAttachNotificationSyncFiles(vmds);
					numberOfTargets = CommonDirectory.getNumberOfTargets();
					int status = CommonDirectory.notifyVm(numberOfTargets);
					// K0532 = status={0}
					if ((IPC.JNI_OK != status)
							&& (CommonDirectory.J9PORT_INFO_SHSEM_OPENED_STALE != status)) {
						throw new AttachNotSupportedException(com.ibm.oti.util.Msg.getString("K0532", status)); //$NON-NLS-1$
					}
				}

				try {
					IPC.logMessage("attachTarget " + targetId + " on port " + portNumber); //$NON-NLS-1$ //$NON-NLS-2$
					targetServer.setSoTimeout(ATTACH_TIMEOUT);
					targetSocket = targetServer.accept();
				} catch (SocketTimeoutException e) {
					targetServer.close();
					IPC.logMessage("attachTarget SocketTimeoutException on " + portNumber + " to " + targetId); //$NON-NLS-1$ //$NON-NLS-2$
					/* [MSG "K0539","acknowledgement timeout from {0} on port {1}"]*/
					throw new AttachNotSupportedException(com.ibm.oti.util.Msg.getString("K0539", targetId, portNumber)); //$NON-NLS-1$
				}
				commandStream = targetSocket.getOutputStream();
				targetSocket.setSoTimeout(COMMAND_TIMEOUT);
				responseStream = targetSocket.getInputStream();
				String response = AttachmentConnection.streamReceiveString(responseStream);
				// K0533 = key error: {0}
				if (!response.contains(' ' + key + ' ')) {
					throw new AttachNotSupportedException(com.ibm.oti.util.Msg.getString("K0533", response)); //$NON-NLS-1$
				}
				IPC.logMessage("attachTarget connected on ", portNumber.toString()); //$NON-NLS-1$
				targetAttached = true;
			} finally {
				if (null != replyFile) {
					replyFile.deleteReply();
				}
				unlockAllAttachNotificationSyncFiles();
				CommonDirectory.cancelNotify(numberOfTargets);
				if (numberOfTargets > 2) {
					try {
						int delayTime = 100 * ((numberOfTargets > 10) ? 10
								: numberOfTargets);
						IPC.logMessage("attachTarget sleep for ", delayTime); //$NON-NLS-1$
						Thread.sleep(delayTime);
					} catch (InterruptedException e) {
						IPC.logMessage("attachTarget sleep interrupted"); //$NON-NLS-1$
					}
				}
				CommonDirectory.releaseAttachLock();
			}
		}
	}

	private void unlockAllAttachNotificationSyncFiles() {

		if (null != targetLocks) {
			for (int i = 0; i < targetLocks.length; ++i) {
				IPC.logMessage("unlockAllAttachNotificationSyncFiles unlocking targetLocks[", i, "]"); //$NON-NLS-1$ //$NON-NLS-2$
				if (null != targetLocks[i]) {
					targetLocks[i].unlockFile();
				}
			}
		}
	}

}
